import 'package:cached_network_image/cached_network_image.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:flutter_webrtc/flutter_webrtc.dart';
import 'package:fluttertoast/fluttertoast.dart';
import 'package:transmedika_one_to_one_call/webrtc/one_to_one/models/profile.dart';

import '../../../websocket/state/socket_state.dart';
import '../bloc/one_to_one_bloc.dart';
import '../bloc/one_to_one_event.dart';
import '../bloc/one_to_one_state.dart';
import '../usecase/one_to_one_usecase.dart';


class OneToOneScreen extends StatelessWidget{
  const OneToOneScreen({
    super.key,
    required this.consultationId,
    required this.mySelf,
    required this.other
  });

  final String consultationId;
  final Profile mySelf;
  final Profile other;
  @override
  Widget build(BuildContext context) => RepositoryProvider.value(
      value: OneToOneUseCase(),
      child: BlocProvider<OneToOneBloc>(
          create: (context) =>
              OneToOneBloc(
                  oneToOneUseCaseService: context.read<OneToOneUseCase>()
              ),
          child: OneToOnePage(consultationId: consultationId, mySelf: mySelf, other: other)
      )
  );
}

class OneToOnePage extends StatefulWidget {
  const OneToOnePage({
    super.key,
    required this.consultationId,
    required this.mySelf,
    required this.other
  });
  final String consultationId;
  final Profile mySelf;
  final Profile other;
  @override
  State<OneToOnePage> createState() => _OneToOnePageState();
}

class _OneToOnePageState extends State<OneToOnePage> {

  final RTCVideoRenderer _localRenderer = RTCVideoRenderer();
  final RTCVideoRenderer _remoteRenderer = RTCVideoRenderer();

  _onConnectToSocket({
    required String url,
    required String from,
    required String to,
    required bool isCalling,
    required Profile mySelf,
    required Profile other,
    required String consultationId
  }) async {
    context
        .read<OneToOneBloc>()
        .add(OnConnectToSocket(
          host: url,
          from: from,
          to: to,
          isCalling: isCalling,
          mySelf: mySelf,
          other: other,
          consultationId: consultationId
      )
    );
  }

  /*_onCloseSocket() async {
    context
        .read<OneToOneBloc>()
        .add(OnSocketClose(ex: "Close Socket from UI", from: true));
  }*/

  _onCloseSession() async{
    context
        .read<OneToOneBloc>()
        .add(OnCloseSession());
  }

  _onCallAcceptedByUser() async{
    context
        .read<OneToOneBloc>()
        .add(OnCallAccepted());
  }

  _onCallRejectedByUser() async{
    context
        .read<OneToOneBloc>()
        .add(OnCallRejected());
  }

  _onStopCall() async{
    context
        .read<OneToOneBloc>()
        .add(OnStopCall());
  }

  _onMute() async{
    context
        .read<OneToOneBloc>()
        .add(OnMute());
  }

  _onSwitchCamera() async{
    context
        .read<OneToOneBloc>()
        .add(OnSwitchCamera());
  }

  _onEnableVideo() async{
    context
        .read<OneToOneBloc>()
        .add(OnEnableVideo());
  }

  _initRenderers() async {
    await _localRenderer.initialize();
    await _remoteRenderer.initialize();
  }

  _disposeRenderers() async {
    _localRenderer.srcObject = null;
    _remoteRenderer.srcObject = null;
    await _localRenderer.dispose();
    await _remoteRenderer.dispose();
  }

  @override
  void initState() {
    _initRenderers();
    _onConnectToSocket(
        url: "wss://rtc.mybeam.me/one2one",
        from: widget.mySelf.email,
        to: widget.other.email,
        isCalling: false,
        mySelf: widget.mySelf,
        other: widget.other,
        consultationId: widget.consultationId
    );
    super.initState();
  }

  @override
  void dispose() {
    _disposeRenderers();
    //_onCloseSocket();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
        body: BlocConsumer<OneToOneBloc, OneToOneUiState>(
            listenWhen: (previous, current) {
              return previous!=current;
            },
            listener: (context, state) {
              debugPrint('State: ${state.socketState?.socketConnectionState.toString()}');
              switch(state.socketState?.socketConnectionState){
                case SocketConnectionState.connecting :
                  break;
                case SocketConnectionState.open :
                  debugPrint('Message: ${state.socketState?.message}');
                  break;
                case SocketConnectionState.close :
                case SocketConnectionState.error :
                  debugPrint('Error: ${state.socketState?.exception}');
                  break;
                default:
              }

              if(state.callState == CallState.missedCall){
                Fluttertoast.showToast(
                    msg: "Missed call from ${state.other?.fullName}",
                    toastLength: Toast.LENGTH_LONG,
                    gravity: ToastGravity.BOTTOM,
                    timeInSecForIosWeb: 1,
                    fontSize: 14.0
                );
              }
            },
            builder: (context, state) {
              String? socketStateStr = "";
              Color socketColor = Colors.white;
              switch(state.socketState?.socketConnectionState){
                case SocketConnectionState.connecting:
                  socketStateStr = "Connecting";
                  socketColor = Colors.yellow;
                  break;
                case SocketConnectionState.open:
                  socketStateStr = "Connected";
                  socketColor = Colors.green;
                  break;
                case SocketConnectionState.idle:
                  socketStateStr = "Idle";
                  socketColor = Colors.blue;
                  break;
                case SocketConnectionState.error:
                case SocketConnectionState.close:
                  socketStateStr = state.socketState?.exception;
                  socketColor = Colors.red;
                  break;
                default:
              }

              String? callStateStr = "";
              Color callColor = Colors.white;
              switch(state.callState){
                case CallState.registering:
                  callStateStr = "Registering";
                  callColor = Colors.yellow;
                  break;
                case CallState.registerAccepted:
                  callStateStr = "Register Accepted and Notify Caller";
                  callColor = Colors.green;
                  break;
                case CallState.registerRejected:
                  callStateStr = "Register Rejected";
                  callColor = Colors.red;
                  break;
                case CallState.callAccepted:
                  callStateStr = "Call Accepted";
                  callColor = Colors.green;
                  break;
                case CallState.callRejected:
                  callStateStr = "Call Rejected";
                  callColor = Colors.red;
                  _onCloseSession();
                  break;
                case CallState.incomingCall:
                  callStateStr = "Incoming Call From: ${state.socketState?.messageModel?.from?.name}";
                  callColor = Colors.green;
                  break;
                case CallState.startCommunication:
                  callStateStr = "Start Communication";
                  callColor = Colors.green;
                  break;
                case CallState.iceCandidate:
                  callStateStr = "Ice Candidate/On Call";
                  callColor = Colors.green;
                  break;
                case CallState.stopCommunication:
                  callStateStr = "Stop Communication";
                  callColor = Colors.green;
                  _disposeRenderers();
                  if (Navigator.canPop(context)) {
                    Navigator.pop(context);
                  } else {
                    SystemNavigator.pop();
                  }
                  break;
                case CallState.missedCall:
                  _disposeRenderers();
                  if (Navigator.canPop(context)) {
                    Navigator.pop(context);
                  } else {
                    SystemNavigator.pop();
                  }
                  break;
                default:
              }

              if(state.remoteStream!=null){
                _remoteRenderer.srcObject = state.remoteStream;
              }

              if(state.removeRemoteStream!=null){
                _remoteRenderer.srcObject = null;
              }

              if(state.localStream!=null) {
                _localRenderer.srcObject = state.localStream;
              }

              return OrientationBuilder(builder: (context, orientation) {
                return Stack(children: <Widget>[
                  Positioned(
                      left: 0.0,
                      right: 0.0,
                      top: 0.0,
                      bottom: 0.0,
                      child: Container(
                        margin: const EdgeInsets.fromLTRB(0.0, 0.0, 0.0, 0.0),
                        width: MediaQuery.of(context).size.width,
                        height: MediaQuery.of(context).size.height,
                        decoration: const BoxDecoration(color: Colors.black54),
                        child: RTCVideoView(_remoteRenderer),
                      )
                  ),
                  Positioned(
                    left: 20.0,
                    top: 40.0,
                    child: Visibility(
                      visible: state.localStream!=null,
                      child: Container(
                        width: orientation == Orientation.portrait ? 110.0 : 140.0,
                        height: orientation == Orientation.portrait ? 140.0 : 110.0,
                        decoration: const BoxDecoration(color: Colors.black54),
                        child: RTCVideoView(_localRenderer),
                      ),
                    )
                  ),
                  Align(
                      alignment: Alignment.topRight,
                      child: Visibility(
                        visible: state.localStream!=null,
                        child: Padding(
                            padding: const EdgeInsets.all(16.0),
                            child: Column(
                              crossAxisAlignment: CrossAxisAlignment.start,
                              mainAxisSize: MainAxisSize.min,
                              children: [
                                const SizedBox(height: 20),
                                ElevatedButton(
                                  onPressed: () {
                                    _onSwitchCamera();
                                  },
                                  style: ElevatedButton.styleFrom(
                                    shape: const CircleBorder(),
                                    padding: const EdgeInsets.all(20),
                                    backgroundColor: Colors.blue,
                                    foregroundColor: Colors.red,
                                  ),
                                  child: state.frontCamera == false ? const Icon(Icons.camera_front, color: Colors.white) : const Icon(Icons.camera_rear, color: Colors.white),
                                ),
                                const SizedBox(height: 10),
                                ElevatedButton(
                                  onPressed: () {
                                    _onMute();
                                  },
                                  style: ElevatedButton.styleFrom(
                                    shape: const CircleBorder(),
                                    padding: const EdgeInsets.all(20),
                                    backgroundColor: Colors.blue,
                                    foregroundColor: Colors.red,
                                  ),
                                  child: state.mute == false ? const Icon(Icons.mic_off, color: Colors.white) : const Icon(Icons.mic, color: Colors.white),
                                ),
                                const SizedBox(height: 10),
                                ElevatedButton(
                                  onPressed: () {
                                    _onEnableVideo();
                                  },
                                  style: ElevatedButton.styleFrom(
                                    shape: const CircleBorder(),
                                    padding: const EdgeInsets.all(20),
                                    backgroundColor: Colors.blue,
                                    foregroundColor: Colors.red,
                                  ),
                                  child: state.enableVideo == true ? const Icon(Icons.videocam_off, color: Colors.white) : const Icon(Icons.videocam, color: Colors.white),
                                )
                              ],
                            )
                        )
                      )
                  ),
                  Align(
                    alignment: orientation == Orientation.portrait ? Alignment.bottomCenter: Alignment.bottomRight,
                    child: Padding(
                        padding: orientation == Orientation.portrait ? const EdgeInsets.only(bottom: 80) : const EdgeInsets.only(bottom: 16, right: 16),
                        child: Row(
                          crossAxisAlignment: CrossAxisAlignment.start,
                          mainAxisSize: MainAxisSize.min,
                          children: [
                            Visibility(
                              visible: state.callState == CallState.incomingCall ||
                                  state.callState == CallState.startCommunication ||
                                  state.callState == CallState.iceCandidate,
                              child: ElevatedButton(
                                onPressed: () {
                                  if(state.callState == CallState.incomingCall){
                                    _onCallRejectedByUser();
                                  }else{
                                    _onStopCall();
                                  }
                                  _disposeRenderers();
                                  if (Navigator.canPop(context)) {
                                    Navigator.pop(context);
                                  } else {
                                    SystemNavigator.pop();
                                  }
                                },
                                style: ElevatedButton.styleFrom(
                                  shape: const CircleBorder(),
                                  padding: const EdgeInsets.all(20),
                                  backgroundColor: Colors.red,
                                  foregroundColor: Colors.red,
                                ),
                                child: const Icon(Icons.call_end, color: Colors.white),
                              ),
                            ),
                            Visibility(
                              visible: state.callState == CallState.incomingCall,
                              child: orientation == Orientation.portrait ? const SizedBox(width: 50) : const SizedBox(width: 30),
                            ),
                            Visibility(
                              visible: state.callState == CallState.incomingCall,
                              child: ElevatedButton(
                                onPressed: () {
                                  _onCallAcceptedByUser();
                                },
                                style: ElevatedButton.styleFrom(
                                  shape: const CircleBorder(),
                                  padding: const EdgeInsets.all(20),
                                  backgroundColor: Colors.green,
                                  foregroundColor: Colors.green,
                                ),
                                child: const Icon(Icons.call, color: Colors.white),
                              ),
                            )
                          ],
                        )
                    )
                  ),
                  Align(
                    alignment: Alignment.topCenter,
                    child: Visibility(
                      visible: state.callState == null || state.callState == CallState.incomingCall ||
                          state.callState == CallState.registering || state.callState == CallState.registerAccepted,
                      child: Padding(
                          padding: const EdgeInsets.only(top:64, right: 32, left: 32),
                          child: Column(
                            crossAxisAlignment: CrossAxisAlignment.center,
                            mainAxisSize: MainAxisSize.min,
                            children: [
                              ClipRRect(
                                borderRadius: BorderRadius.circular(120),
                                child: CachedNetworkImage(
                                  width: 190,
                                  height: 190,
                                  imageUrl: state.other?.profilePicture ?? '',
                                  imageBuilder: (context, imageProvider) => Container(
                                    decoration: BoxDecoration(
                                      image: DecorationImage(
                                          image: imageProvider,
                                          fit: BoxFit.cover),
                                    ),
                                  ),
                                  placeholder: (context, url) => const Icon(
                                    Icons.account_circle_outlined,
                                    color: Colors.white,
                                    size: 190,
                                  ),
                                  errorWidget: (context, url, error) => const Icon(
                                    Icons.account_circle_outlined,
                                    color: Colors.red,
                                    size: 190,
                                  ),
                                ),
                              ),
                              const SizedBox(height: 30,),
                              Container(
                                padding: const EdgeInsets.only(right: 80.0, left: 80),
                                child: Text(state.other?.fullName ?? '-',
                                  overflow: TextOverflow.ellipsis,
                                  maxLines: 1,
                                  style: const TextStyle(
                                      color: Colors.white,
                                      fontSize: 20
                                  ),
                                ),
                              ),
                            ],
                          )
                      ),
                    )
                  ),
                  Align(
                      alignment: Alignment.bottomLeft,
                      child: Padding(
                          padding: const EdgeInsets.all(16.0),
                          child: Column(
                            crossAxisAlignment: CrossAxisAlignment.start,
                            mainAxisSize: MainAxisSize.min,
                            children: [
                              RichText(
                                text: TextSpan(
                                  text: 'Socket State: ',
                                  style: const TextStyle(color: Colors.white),
                                  children: <TextSpan>[
                                    TextSpan(text: socketStateStr!,
                                        style: TextStyle(
                                            fontWeight: FontWeight.bold,
                                            color: socketColor
                                        )),
                                  ],
                                ),
                              ),
                              RichText(
                                text: TextSpan(
                                  text: 'Call State: ',
                                  style: const TextStyle(color: Colors.white),
                                  children: <TextSpan>[
                                    TextSpan(text: callStateStr!,
                                        style: TextStyle(
                                            fontWeight: FontWeight.bold,
                                            color: callColor
                                        )),
                                  ],
                                ),
                              ),
                            ],
                          )
                      )
                  ),
                ]);
              });
            }
        )
    );
  }
}